from openapi_core.wrappers.base import BaseOpenAPIRequest
from starlette.requests import Request
from starlette.routing import Match


class StarletteOpenAPIRequest(BaseOpenAPIRequest):

    def __init__(self, request: Request):
        self.request = request
        self.body = b''

    @classmethod
    async def prepare(cls, request):
        req = cls(request)
        req.body = await req.request.body()
        return req

    @property
    def host_url(self):
        url = f"{self.request.url.scheme}://{self.request.url.hostname}"
        if self.request.url.port:
            url = f"{url}:{self.request.url.port}"
        return url

    @property
    def path(self):
        return self.request['path']

    @property
    def method(self):
        return self.request.method.lower()

    @property
    def path_pattern(self):
        for route in self.request.app.router.routes:
            match, _ = route.matches(self.request)
            if match == Match.FULL:
                return route.path
        return self.path

    @property
    def parameters(self):
        return {
            'path': self.request.path_params,
            'query': self.request.query_params,
            'header': self.request.headers,
            'cookie': self.request.cookies,
        }

    @property
    def mimetype(self):
        return self.request.headers['content-type']



from typing import Any

from typing_extensions import Protocol


class Requestable(Protocol):  # pragma: no cover
    """ Any object (usually a module, class or instance) that implements the
        `request` method compatible with the `requests` library.
    """
    def request(method: str, url: str, **kwargs) -> Any:
        ...



from functools import wraps
from importlib import import_module
from inspect import iscoroutinefunction
from pathlib import Path
from typing import Callable, Union

from openapi_core import create_spec
from openapi_core.schema.specs.models import Spec
from openapi_core.shortcuts import RequestValidator, ResponseValidator
from starlette.applications import Starlette
from starlette.middleware.base import BaseHTTPMiddleware, RequestResponseEndpoint
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.types import ASGIApp
from stringcase import snakecase

from pyotr.utils import get_spec_from_file
from pyotr.validation.requests import StarletteOpenAPIRequest
from pyotr.validation.responses import StarletteOpenAPIResponse


class Application(Starlette):

    def __init__(
        self, spec: Union[Spec, dict], base: str, *, validate_responses: bool = True, **kwargs
    ):
        super().__init__(**kwargs)
        if not isinstance(spec, Spec):
            spec = create_spec(spec)
        self.spec = spec
        self.validate_responses = validate_responses

        for path, path_spec in spec.paths.items():
            for method, operation in path_spec.operations.items():
                endpoint = self._get_endpoint(operation.operation_id, base)
                self.add_route(path, endpoint, [method])

    def _get_endpoint(self, name: str, base: str, enforce_case: bool = True) -> Callable:
        if '.' in name:
            base, name = f"{base}.{name}".rsplit('.', 1)
        if enforce_case:
            name = snakecase(name)
        try:
            module = import_module(base)
        except ModuleNotFoundError:
            raise RuntimeError(f'The module {base} does not exist!')
        endpoint = getattr(module, name, None)
        if endpoint is None:
            raise RuntimeError(f'The function {base}.{name} does not exist!')

        @wraps(endpoint)
        async def wrapper(request, **kwargs) -> Response:
            validation_request = await StarletteOpenAPIRequest.prepare(request)
            request_validation = RequestValidator(self.spec).validate(validation_request)
            request_validation.raise_for_errors()

            if iscoroutinefunction(endpoint):
                response = await endpoint(request, **kwargs)
            else:
                response = endpoint(request, **kwargs)
            if not isinstance(response, Response):
                response = JSONResponse(response)
            # TODO: a list of endpoint names to specify which responses to skip
            if self.validate_responses:
                ResponseValidator(self.spec).validate(
                    StarletteOpenAPIRequest(request),
                    StarletteOpenAPIResponse(response)
                ).raise_for_errors()
            return response

        return wrapper

    @classmethod
    def from_file(cls, path: Union[Path, str], *args, **kwargs) -> 'Application':
        spec = get_spec_from_file(path)
        return cls(spec, *args, **kwargs)


class OpenAPIRequestValidationMiddleware(BaseHTTPMiddleware):

    def __init__(self, app: ASGIApp, spec: Spec, **kwargs):
        super().__init__(app, **kwargs)
        self.spec = spec

    async def dispatch(self, request: Request, call_next: RequestResponseEndpoint) -> Response:
        request = await StarletteOpenAPIRequest.prepare(request)
        request_validation = RequestValidator(self.spec).validate(request)
        request_validation.raise_for_errors()
        response = await call_next(request.request)
        return response



from openapi_core.wrappers.base import BaseOpenAPIResponse
from starlette.responses import Response


class StarletteOpenAPIResponse(BaseOpenAPIResponse):

    def __init__(self, response: Response):
        self.body = response.body
        self.status_code = response.status_code
        self.mimetype = response.headers.get('content-type')
        self.data = self.body



# -*- coding: utf-8 -*-
from distutils.core import setup

package_dir = \
{'': 'src'}

packages = \
['pyotr', 'pyotr.validation.requests', 'pyotr.validation.responses']

package_data = \
{'': ['*']}

install_requires = \
['http3>=0.6.3,<0.7.0',
 'openapi-core>=0.11.0,<0.12.0',
 'pytest-cov>=2.7,<3.0',
 'pyyaml>=5.1,<6.0',
 'requests>=2.22,<3.0',
 'starlette>=0.12.1,<0.13.0',
 'stringcase>=1.2,<2.0',
 'typing-extensions>=3.7,<4.0']

extras_require = \
{'uvicorn': ['uvicorn>=0.8.3,<0.9.0']}

setup_kwargs = {
    'name': 'pyotr',
    'version': '0.1.2',
    'description': 'Python OpenAPI-to-REST (and back) framework ',
    'long_description': 'Pyotr\n=====\n\n**Pyotr** is a Python library for serving and consuming REST APIs based on [OpenAPI](https://swagger.io/resources/open-api/) specifications. Its name is acronym of "Python OpenAPI to REST".\n\nThe project consists of two separate libraries that can be used independently:\n\n* `pyotr.server` is a [Starlette](https://www.starlette.io)-based framework for serving OpenAPI-based services. It is functionally very similar to [connexion](https://connexion.readthedocs.io), except that it aims to be fully [ASGI](https://asgi.readthedocs.io)-compliant. \n* `pyotr.client` is a simple HTTP client for consuming OpenAPI-based services.\n\n**WARNING:** This is still very much work in progress and not nearly ready for any kind of production.\n\n\nQuick Start\n-----------\n\n### Server\n\n    from pyotr.server import Application\n    \n    app = Application.from_file("path/to/openapi.yaml", "path.to.endpoints.module")\n    \n### Client\n\n    from pyotr.client import Client\n    \n    client = Client.from_file("path/to/openapi.yaml")\n    result = client.some_endpoint_id("path", "variables", "query_var"="example")\n    \nWhy Pyotr?\n----------\n\nThe main advantage of Pyotr -- both as a server and as a client -- is that it uses the OpenAPI specification to both prepare and validate requests and responses.\n\nIn `pyotr.server`, the specification is used to route a request to the right endpoint, allowing the developer to simply implement the endpoint function. And for `pyotr.client`, the spec is used to construct the request, applying any arguments provided by the user, before sending it to the server.\n',
    'author': 'Berislav Lopac',
    'author_email': 'berislav@lopac.net',
    'url': 'https://github.com/berislavlopac/pyotr',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
